<?php
declare(strict_types=1);

namespace Drupal\generic_layout;

use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\layout_builder\LayoutEntityHelperTrait;

use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\layout_builder\SectionStorage\SectionStorageManager;
use Drupal\layout_builder\LayoutTempstoreRepository;
use Drupal\Core\File\FileSystemInterface;

use Drupal\generic_layout\Plugin\Layout\GenericLayout;


class GenericLayoutManager {
  use LayoutEntityHelperTrait;
  const DIRECTORY = "public://generic_layout";

  protected $entityTypeManager;
  protected EntityFieldManagerInterface $entityFieldManager;
  protected LayoutTempstoreRepository $layoutTempstoreRepository;
  protected SectionStorageManager $sectionStorage;
  protected FileSystemInterface $fileSystem;


  public function __construct(EntityTypeManagerInterface $entity_type_manager, EntityFieldManagerInterface $entity_field_manager, SectionStorageManager $section_storage, LayoutTempstoreRepository $layout_tempstore_repo, FileSystemInterface $file_system) {
    $this->entityTypeManager = $entity_type_manager;
    $this->entityFieldManager = $entity_field_manager;
    $this->sectionStorage = $section_storage;
    $this->layoutTempstoreRepository = $layout_tempstore_repo;
    $this->fileSystem = $file_system;
  }

  public static function getCssUri(GenericLayout $layout): string {
    $subfolder = $layout->getPluginDefinition()->id();
    $uri = "public://{$subfolder}/{$layout->getInstance()}.css";

    return $uri;
  }

  private function getLayoutBuilderEntityMap(): array {
    $map = [];

    foreach ($this->entityFieldManager->getFieldMapByFieldType('layout_section') as $entity_type_id => $data) {
      $sections_field_name = array_key_first($data);

      $entityStorage = $this->entityTypeManager->getStorage($entity_type_id);
      $query = $entityStorage->getQuery()->accessCheck(FALSE);

      // @quickfix
      if($entity_type_id == 'node') {
        $query->condition('type', 'article');
      }

      $map[$entity_type_id] = $query->execute();
    }

    return $map;
  }

  public function getGenericLayoutSections(): array {
    $sections = [];

    foreach ($this->getLayoutBuilderEntityMap() as $entity_type_id => $ids) {
      $entityStorage = $this->entityTypeManager->getStorage($entity_type_id);


      // @fixme only fetch the view_modes, where layout_builder is used!
      // $view_modes = \Drupal::service('entity_display.repository')->getViewModes($entity_type_id) ?: ['full', 'default'];
      $view_modes = ['full', 'default'];

      foreach ($view_modes as $view_mode) {
        foreach ($entityStorage->loadMultiple($ids) as $entity) {
          $sectionStorage = $this->getSectionStorageForEntity($entity, $view_mode);
          $sections = array_merge($sections, $sectionStorage->getSections());

          if ($this->layoutTempstoreRepository->has($sectionStorage)) {
            $sectionStorageTemp = $this->layoutTempstoreRepository->get($sectionStorage);
            $sections = array_merge($sections, $sectionStorageTemp->getSections());
          }
        }
      }
    }

    $gl_sections = GenericLayoutHelpers::extractGenericLayoutSections($sections);
    return $gl_sections;
  }

  // @see: https://github.com/sabberworm/PHP-CSS-Parser#use-parser-to-prepend-an-id-to-all-selectors
  public static function prefixSelectors(string $prefix, \Sabberworm\CSS\CSSList\Document &$css) {
    foreach ($css->getAllDeclarationBlocks() as $block) {
      foreach ($block->getSelectors() as $selector) {
        $selector->setSelector($prefix . ' ' . $selector->getSelector());
      }
    }
  }

  public function storeCssFile(GenericLayout $layout): string {
    $instance = $layout->getInstance();
    $css_uri = self::getCssUri($layout);

    $css = $layout->getCss();
    self::prefixSelectors('.' . $instance, $css);
    $rendered_css = $css->render(\Sabberworm\CSS\OutputFormat::createPretty());

    $directory = self::DIRECTORY;
    $this->fileSystem->prepareDirectory($directory, FileSystemInterface::CREATE_DIRECTORY | FileSystemInterface::MODIFY_PERMISSIONS);

    return $this->fileSystem->saveData($rendered_css, $css_uri, FileSystemInterface::EXISTS_REPLACE);
  }

  public function removeCssFile(GenericLayout $layout): void {
    $css_uri = self::getCssUri($layout);
    $this->fileSystem->delete($css_uri);
  }

  public function removeAbandonedCssFiles(array $files_to_keep): void {
    $files = $this->fileSystem->scanDirectory(self::DIRECTORY, '/.*/', ['key' => 'name']);
    $abandoned_files = array_diff_key($files, $files_to_keep);

    foreach($abandoned_files as $uri => $data) {
      $this->fileSystem->delete($data->uri);
    }
  }

  public function removeCssFiles() {
    // @todo which fits best?
    // $this->fileSystem->rmdir(self::DIRECTORY);
    $this->fileSystem->deleteRecursive(self::DIRECTORY);

    \Drupal::service('library.discovery')->clearCachedDefinitions();
  }

  public function getLibrary(GenericLayout $layout): array {
    $css_uri = self::getCssUri($layout);

    if (!file_exists($css_uri)) {
      $this->storeCssFile($layout);
    }

    return [
      'css' => [
        'layout' => [
          $css_uri => [
            'preprocess' => false,
          ]
        ]
      ]
    ];
  }

  public function getLibraries(): array {
    $libraries = [];

    foreach ($this->getGenericLayoutSections() as $section) {
      $layout = $section->getLayout();
      $libraries[$layout->getInstance()] = self::getLibrary($layout);
    }

    // Now is a cheap and reliable moment to get rid of abandoned css files
    $this->removeAbandonedCssFiles($libraries);

    return $libraries;
  }

  /**
   * If the LayoutSettings scheme changes, this helper makes it easy, to update existing Layouts accordingly.
   * @todo due to the need of changing all Settings in normal operation (f.e. when changing the breakpoint group),
   * the function may be moved to GridLayoutManager
   * This is typically followed by a call of clearLayoutTempstoreRepositories() to also get rid of invalid configurations
   * in currenlty edited layouts
   *
   * Example updpate converting each GenericLayout's key 'uuid' to 'instance':
   *
   * \Drupal::service('generic_layout.manager')->updateLayoutSettings( function(&$settings) {
   *   if(!isset($settings['instance'])) {
   *     $settings['instance'] = $settings['uuid'];
   *   }
   *   unset($settings['uuid']);
   * });
   *
   * @param $callback
   * @return void
   */
  public function updateLayoutSettings(callable $callable, bool $clear_cache = TRUE): void {
    $sections = [];
    foreach ($this->getLayoutBuilderEntityMap() as $entity_type_id => $ids) {
      $entityStorage = $this->entityTypeManager->getStorage($entity_type_id);

      foreach ($entityStorage->loadMultiple($ids) as $entity) {
        $sections = $this->getSectionStorageForEntity($entity)->getSections();
        $gl_sections = GenericLayoutHelpers::extractGenericLayoutSections($sections);

        foreach ($gl_sections as $section) {
          $settings = $section->getLayoutSettings();

          // Invoke user function
          // @todo use 'Drupal Batch API' cause a site may have dozens of Layouts
          // @see https://www.drupal.org/docs/drupal-apis/batch-api/batch-api-overview
          call_user_func_array($callable, [&$settings]);

          $section->setLayoutSettings($settings);
        }
        $entity->save();
      }
    }

    if ($clear_cache) {
      \Drupal::service('library.discovery.collector')->delete('generic_layout');
    }
  }

  /**
   * Deletes all layouts currently edited
   */
  public function clearLayoutTempstoreRepositories(): int {
    $tempstores_deleted = 0;

    foreach ($this->getLayoutBuilderEntityMap() as $entity_type_id => $ids) {
      $entityStorage = $this->entityTypeManager->getStorage($entity_type_id);
      foreach ($entityStorage->loadMultiple($ids) as $entity) {
        // @fixme take all view_modes into account
        $view_mode = 'default';
        $sectionStorage = $this->getSectionStorageForEntity($entity, $view_mode);

        if ($this->layoutTempstoreRepository->has($sectionStorage)) {
          $this->layoutTempstoreRepository->delete($sectionStorage);
          $tempstores_deleted++;
        }
      }
    }

    return $tempstores_deleted;
  }
}
